<?php

    /** @noinspection PhpMissingFieldTypeInspection */

    namespace TamerLib\Classes;

    use Exception;
    use LogLib\Log;
    use Redis;
    use TamerLib\Enums\JobStatus;
    use TamerLib\Exceptions\JobNotFoundException;
    use TamerLib\Exceptions\ConnectionException;
    use TamerLib\Exceptions\ServerException;
    use TamerLib\Exceptions\TimeoutException;
    use TamerLib\Objects\JobPacket;
    use TamerLib\Objects\ServerConfiguration;
    use TamerLib\Objects\WorkerConfiguration;
    use Throwable;

    class JobManager
    {
        /**
         * @var ServerConfiguration
         */
        private $server_configuration;

        /**
         * @var Redis
         */
        private $redis_client;

        /**
         * @var int|null
         */
        private $last_connect;

        /**
         * JobManager constructor.
         *
         * @param ServerConfiguration $serverConfiguration
         */
        public function __construct(ServerConfiguration $serverConfiguration)
        {
            $this->server_configuration = $serverConfiguration;
            $this->redis_client = new Redis();
        }

        /**
         * Attempts to determine if the Redis Server is online.
         *
         * @param bool $ping
         * @return bool
         */
        private function isConnected(bool $ping=true): bool
        {
            if($this->redis_client === null)
            {
                return false;
            }

            try
            {
                if($ping)
                {
                    $this->redis_client->ping();
                }
            }
            catch(Exception $e)
            {
                unset($e);
                return false;
            }

            return true;
        }

        /**
         * Attempts to connect to the Redis Server
         *
         * @return void
         * @throws ConnectionException
         */
        private function connect(): void
        {
            // Reconnect every 30 minutes
            if ($this->last_connect !== null && $this->last_connect < (time() - 1800))
            {
                //$this->disconnect();
            }

            if($this->isConnected())
            {
                return;
            }

            try
            {
                $this->redis_client->connect(
                    $this->server_configuration->getHost(),
                    $this->server_configuration->getPort()
                );

                if ($this->server_configuration->getPassword() !== null)
                {
                    $this->redis_client->auth($this->server_configuration->getPassword());
                }

                $this->redis_client->select($this->server_configuration->getDatabase());
                $this->last_connect = time();
            }
            catch(Exception $e)
            {
                throw new ConnectionException(sprintf('Could not connect to %s:%s', $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }
        }

        /**
         * Disconnects from the Redis Server
         *
         * @return void
         */
        private function disconnect(): void
        {
            try
            {
                $this->redis_client->close();
            }
            catch(Exception $e)
            {
                Log::warning('net.nosial.tamerlib',sprintf('JobManager could not disconnect safely from %s:%s', $this->server_configuration->getHost(), $this->server_configuration->getPort()), $e);
            }
            finally
            {
                $this->redis_client = null;
            }
        }

        /**
         * Returns the Redis Client, or attempts to connect to the Redis Server if not connected.
         *
         * @return Redis
         * @throws ConnectionException
         */
        private function getClient(): Redis
        {
            $this->connect();
            return $this->redis_client;
        }

        /**
         * Pushes a JobPacket to the Redis server.
         *
         * @param JobPacket $jobPacket
         * @return void
         * @throws ServerException
         */
        public function pushJob(JobPacket $jobPacket): void
        {
            try
            {
                $this->getClient()->hMSet($jobPacket->getId(), $jobPacket->toArray());
                $this->getClient()->rPush(sprintf('ch%s', $jobPacket->getChannel()), $jobPacket->getId());
            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not push job to %s:%s', $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }
        }

        /**
         * Attempts to get a specific job from the Redis Server.
         *
         * @param JobPacket|string $job_id
         * @return JobPacket
         * @throws JobNotFoundException
         * @throws ServerException
         */
        public function getJob(JobPacket|string $job_id): JobPacket
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist in %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                return new JobPacket($this->getClient()->hGetAll($job_id));

            }
            catch(JobNotFoundException $e)
            {
                throw $e;
            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not get job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }
        }

        /**
         * Returns the status of a job.
         *
         * @param JobPacket|string $job_id
         * @return int
         * @throws ServerException
         */
        public function getJobStatus(JobPacket|string $job_id): int
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                return (int)$this->getClient()->hGet($job_id, 'status');
            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not get job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }
        }

        /**
         * Listens on a return channel for a returned job.
         *
         * @param string $return_channel
         * @param int $timeout
         * @return JobPacket
         * @throws JobNotFoundException
         * @throws ServerException
         * @throws TimeoutException
         */
        public function listenReturnChannel(string $return_channel, int $timeout=0): JobPacket
        {
            $time_start = time();

            try
            {
                if($timeout < 0)
                {
                    Log::debug('net.nosial.tamerlib', sprintf('Listening for job on return channel %s on %s:%s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                    $job_packet = $this->getClient()->lPop($return_channel);
                }
                else
                {
                    Log::debug('net.nosial.tamerlib', sprintf('Listening for job on return channel %s on %s:%s with timeout %s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort(), $timeout));
                    $job_packet = $this->getClient()->blPop($return_channel, $timeout);
                }
            }
            catch(Exception $e)
            {
                throw new TimeoutException(sprintf('Could not get job from return channel %s on %s:%s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }

            if($job_packet === null)
            {
                throw new TimeoutException(sprintf('Could not get job from return channel %s on %s:%s', $return_channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
            }

            if($job_packet[1] === 'test')
            {
                return $this->listenReturnChannel($return_channel, $timeout);
            }

            Log::debug('net.nosial.tamerlib', sprintf('Received job %s from return channel %ss', $job_packet[1], $return_channel));
            return $this->getJob($job_packet[1]);
        }

        /**
         * Gets the return value of a job.
         *
         * @param JobPacket|string $job_id
         * @return mixed
         * @throws JobNotFoundException
         * @throws ServerException
         */
        public function getJobResult(JobPacket|string $job_id): mixed
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                return $this->getClient()->hGet($job_id, 'return');
            }
            catch(JobNotFoundException $e)
            {
                throw $e;
            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not get job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }
        }

        /**
         * Returns the exception to a job.
         *
         * @param JobPacket|string $job_id
         * @return Throwable
         * @throws ServerException
         */
        public function getJobException(JobPacket|string $job_id): Throwable
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            try
            {
                if(!$this->getClient()->exists($job_id))
                {
                    throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                }

                return unserialize($this->getClient()->hGet($job_id, 'exception'), ['allowed_classes'=>true]);
            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not get job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }
        }

        /**
         * Attempts to claim an existing JobPacket, returns True if the worker's claim was successful.
         * Returns False if the JobPacket does not exist, or if the JobPacket is already claimed.
         *
         * @param JobPacket|string $job_id
         * @param WorkerConfiguration|string $worker_id
         * @return bool
         * @throws ServerException
         */
        public function claimJob(JobPacket|string $job_id, WorkerConfiguration|string $worker_id): bool
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            if($worker_id instanceof WorkerConfiguration)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $worker_id = $worker_id->getWorkerId();
            }

            try
            {
                // Check if the job exists
                if(!$this->getClient()->exists($job_id))
                {
                    return false;
                }

                // Attempt to claim the job
                $this->getClient()->hSet($job_id, 'worker_id', $worker_id);
                // Verify that the job was claimed
                if($this->getClient()->hGet($job_id, 'worker_id') !== $worker_id)
                {
                    return false;
                }

                // Set the job status to processing
                $this->getClient()->hSet($job_id, 'status', JobStatus::PROCESSING);
            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not claim job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }

            return true;
        }

        /**
         * Marks a job as finished, and sets the return value.
         *
         * @param JobPacket|string $job_id
         * @param mixed|null $return_value
         * @return void
         * @throws JobNotFoundException
         * @throws ServerException
         */
        public function returnJob(JobPacket|string $job_id, mixed $return_value=null): void
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            try
            {
                if($this->getClient()->exists($job_id))
                {
                    $return_channel = $this->getClient()->hGet($job_id, 'return_channel');
                    Log::debug('net.nosial.tamerlib', sprintf('Returning job %s (Return Channel: %s)', $job_id, $return_channel ?? 'n/a'));

                    if($return_channel === null)
                    {
                        $this->getClient()->del($job_id);
                        return;
                    }

                    $this->getClient()->hSet($job_id, 'return_value', serialize($return_value));
                    $this->getClient()->hSet($job_id, 'status', JobStatus::FINISHED);
                    $this->getClient()->rPush($this->getClient()->hGet($job_id, 'return_channel'), $job_id);
                    return;
                }
            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not return job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }

            throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
        }

        /**
         * Rejects a job and pushes it back to the channel.
         *
         * @param JobPacket|string $job_id
         * @return void
         * @throws JobNotFoundException
         * @throws ServerException
         */
        public function rejectJob(JobPacket|string $job_id): void
        {
            $channel_id = null;

            if($job_id instanceof JobPacket)
            {
                // Providing a JobPacket allows us to get the channel_id and avoid an extra call to the server
                $channel_id = $job_id->getChannel();
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            try
            {
                if($this->getClient()->exists($job_id))
                {
                    Log::debug('net.nosial.tamerlib', sprintf('Rejecting job %s', $job_id));

                    // Mark as rejected, clear worker_id
                    $this->getClient()->hSet($job_id, 'worker_id', null);
                    // Push back to the channel
                    if($channel_id !== null)
                    {
                        $channel_id = $this->getClient()->hGet($job_id, 'channel');
                    }
                    $this->getClient()->rPush(sprintf('ch%s', $channel_id), $job_id);

                    return;
                }
            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not reject job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }

            throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
        }

        /**
         * Sets the job as failed, and sets the exception that was thrown.
         *
         * @param JobPacket|string $job_id
         * @param Throwable $throwable
         * @return void
         * @throws ServerException
         */
        public function returnException(JobPacket|string $job_id, Throwable $throwable): void
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            try
            {
                if($this->getClient()->exists($job_id))
                {
                    Log::debug('net.nosial.tamerlib', sprintf('Returning exception for job %s', $job_id));

                    if($this->getClient()->hGet($job_id, 'return_channel') === null)
                    {
                        $this->getClient()->del($job_id);
                        return;
                    }

                    $this->getClient()->hSet($job_id, 'exception', serialize($throwable));
                    $this->getClient()->hSet($job_id, 'status', JobStatus::FAILED);
                    $this->getClient()->rPush($this->getClient()->hGet($job_id, 'return_channel'), $job_id);
                    Log::debug('net.nosial.tamerlib', sprintf('Pushed job %s to return channel %s', $job_id, $this->getClient()->hGet($job_id, 'return_channel')));
                    return;
                }
            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not return exception for job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }
        }

        /**
         * Waits for a job to be available on the channel or multiple channels, attempts to claim it, and returns it.
         *
         * If multiple channels are selected, the function will iterate through each channel until a job is found.
         *
         * @param WorkerConfiguration|string $worker_id The worker that is waiting for the job.
         * @param int|array $channel The channel (or channels) to wait for the job on.
         * @param int $timeout The timeout in seconds to wait for the job.
         * @return JobPacket The job that was claimed, or null if no job was available.
         * @throws ServerException
         * @throws TimeoutException
         */
        public function listenForJob(WorkerConfiguration|string $worker_id, int|array $channel=0, int $timeout=0): JobPacket
        {
            if($worker_id instanceof WorkerConfiguration)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $worker_id = $worker_id->getWorkerId();
            }

            $channels = [];
            if(is_array($channel))
            {
                foreach($channel as $id)
                {
                    $channels[] = sprintf('ch%s', $id);
                }
            }
            else
            {
                $channels[] = sprintf('ch%s', $channel);
            }

            try
            {
                while(true)
                {
                    if($timeout < 0)
                    {
                        $job_id = $this->getClient()->blPop($channels, 1);
                    }
                    else
                    {
                        $job_id = $this->getClient()->blPop($channels, $timeout);
                    }

                    if($job_id !== false && $this->claimJob($job_id[1], $worker_id))
                    {
                        return new JobPacket($this->getClient()->hGetAll($job_id[1]));
                    }

                    if($timeout < 0)
                    {
                        throw new TimeoutException(sprintf('Timeout exceeded while waiting for job on %s:%s', $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                    }

                    if($timeout === 0 && $job_id === null)
                    {
                        throw new TimeoutException(sprintf('Timeout exceeded while waiting for job on %s:%s', $this->server_configuration->getHost(), $this->server_configuration->getPort()));
                    }
                }

            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not wait for job on channel %s from %s:%s', $channel, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }
        }

        /**
         * Drops a job from the server entirely.
         *
         * @param JobPacket|string $job_id
         * @return void
         * @throws JobNotFoundException
         * @throws ServerException
         */
        public function dropJob(JobPacket|string $job_id): void
        {
            if($job_id instanceof JobPacket)
            {
                /** @noinspection CallableParameterUseCaseInTypeContextInspection */
                $job_id = $job_id->getId();
            }

            try
            {
                if($this->getClient()->exists($job_id))
                {
                    $this->getClient()->del($job_id);
                    return;
                }
            }
            catch(Exception $e)
            {
                throw new ServerException(sprintf('Could not drop job %s from %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()), 0, $e);
            }

            throw new JobNotFoundException(sprintf('Job %s does not exist on %s:%s', $job_id, $this->server_configuration->getHost(), $this->server_configuration->getPort()));
        }
    }